@@ -29,7 +29,7 @@
 ** decode strings in HTML or HTTP.
 */
 #include <config.h>
-#include "cgi.h"
+#include <ctype.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -37,6 +37,11 @@
 #include <sys/times.h>
 #include <sys/time.h>
 #include <sys/wait.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <sys/select.h>
+#include <unistd.h>
+#include "cgi.h"
 
 #if INTERFACE
 /*
@@ -55,8 +60,8 @@
 ** Provide a reliable implementation of a caseless string comparison
 ** function.
 */
-#define stricmp sqliteStrICmp
-extern int sqliteStrICmp(const char*, const char*);
+#define stricmp sqlite3StrICmp
+extern int sqlite3StrICmp(const char*, const char*);
 
 /*
 ** The body of the HTTP reply text is stored here.
@@ -69,7 +74,7 @@ static char *zTxt = 0;    /* Pointer to malloced space */
 ** Append reply content to what already exists.  Return a pointer
 ** to the start of the appended text.
 */
-char *cgi_append_content(const char *zData, int nAmt){
+const char *cgi_append_content(const char *zData, int nAmt){
   if( nUsedTxt+nAmt >= nAllocTxt ){
     nAllocTxt = nUsedTxt*2 + nAmt + 1000;
     zTxt = realloc( zTxt, nAllocTxt );
@@ -85,6 +90,7 @@ char *cgi_append_content(const char *zData, int nAmt){
 */
 void cgi_reset_content(void){
   nUsedTxt = 0;
+  g.zLinkURL = 0;
 }
 
 /*
@@ -142,6 +148,8 @@ void cgi_append_header(const char *zLine){
 
 /*
 ** Set a cookie.
+**
+** Zero lifetime implies a session cookie.
 */
 void cgi_set_cookie(
   const char *zName,    /* Name of the cookie */
@@ -152,10 +160,11 @@ void cgi_set_cookie(
   char *zCookie;
   if( zPath==0 ) zPath = "/";
   if( lifetime>0 ){
-    zCookie = mprintf("SET-COOKIE: %s=%t; PATH=%s; MAX-AGE=%d;\r\n",
-                      zName, zValue, zPath, lifetime);
+		lifetime += (int)time(0);
+    zCookie = mprintf("Set-Cookie: %s=%t; Path=%s; expires=%s; Version=1\r\n",
+                      zName, zValue, zPath, cgi_rfc822_datestamp(lifetime));
   }else{
-    zCookie = mprintf("SET-COOKIE: %s=%t; PATH=%s;\r\n",
+    zCookie = mprintf("Set-Cookie: %s=%t; Path=%s; Version=1\r\n",
                       zName, zValue, zPath);
   }
   cgi_append_header(zCookie);
@@ -171,6 +180,54 @@ void cgi_logfile(const char *zFile, const char *zArg){
   zLogArg = zArg;
 }
 
+static char *cgi_add_etag(char *zTxt, int nLen){
+  MD5Context ctx;
+  unsigned char digest[16];
+  int i, j;
+  char zETag[64];
+
+  MD5Init(&ctx);
+  MD5Update(&ctx,(unsigned char *)zTxt,(unsigned int)nLen);
+  MD5Final(digest,&ctx);
+  for(j=i=0; i<16; i++,j+=2){
+    bprintf(&zETag[j],sizeof(zETag)-j,"%02x",(int)digest[i]);
+  }
+  cgi_append_header( mprintf("ETag: %s\r\n", zETag) );
+  return strdup(zETag);
+}
+
+/*
+** Do some cache control stuff. First, we generate an ETag and include it in
+** the response headers. Second, we do whatever is necessary to determine if
+** the request was asking about caching and whether we need to send back the
+** response body. If we shouldn't send a body, return non-zero.
+**
+** Currently, we just check the ETag against any If-None-Match header.
+**
+** FIXME: In some cases (attachments, file contents) we could check
+** If-Modified-Since headers and always include Last-Modified in responses.
+*/
+static int check_cache_control(void){
+  /* FIXME: there's some gotchas wth cookies and some headers. */
+  char *zETag = cgi_add_etag(zTxt,nUsedTxt);
+  char *zMatch = getenv("HTTP_IF_NONE_MATCH");
+
+  if( zETag!=0 && zMatch!=0 ) {
+    char *zBuf = strdup(zMatch);
+    if( zBuf!=0 ){
+      char *zTok = 0;
+      char *zPos;
+      for( zTok = strtok_r(zBuf, ",\"",&zPos);
+           zTok && strcasecmp(zTok,zETag);
+           zTok =  strtok_r(0, ",\"",&zPos)){}
+      free(zBuf);
+      if(zTok) return 1;
+    }
+  }
+  
+  return 0;
+}
+
 /*
 ** Do a normal HTTP reply
 */
@@ -180,8 +237,20 @@ void cgi_reply(void){
     iReplyStatus = 200;
     zReplyStatus = "OK";
   }
+
+  if( iReplyStatus==200 && check_cache_control() ) {
+    /* change the status to "unchanged" and we can skip sending the
+    ** actual response body. Obviously we only do this when we _have_ a
+    ** body (code 200).
+    */
+    iReplyStatus = 304;
+    zReplyStatus = "Not Modified";
+  }
+
   if( fullHttpReply ){
     printf("HTTP/1.0 %d %s\r\n", iReplyStatus, zReplyStatus);
+    printf("Date: %s\r\n", cgi_rfc822_datestamp(time(0)));
+    printf("Connection: close\r\n");
   }else{
     printf("Status: %d %s\r\n", iReplyStatus, zReplyStatus);
   }
@@ -189,26 +258,40 @@ void cgi_reply(void){
   if( zExtraHeader ){
     printf("%s", zExtraHeader);
   }
+
+  if( g.isConst ){
+    /* constant means that the input URL will _never_ generate anything
+    ** else. In the case of attachments, the contents won't change because
+    ** an attempt to change them generates a new attachment number. In the
+    ** case of most /getfile calls for specific versions, the only way the
+    ** content changes is if someone breaks the SCM. And if that happens, a
+    ** stale cache is the least of the problem. So we provide an Expires
+    ** header set to a reasonable period (default: one week).
+    */
+    time_t expires = time(0) + atoi(db_config("constant_expires","604800"));
+    printf( "Expires: %s\r\n", cgi_rfc822_datestamp(expires));
+  }
+
+  if( g.isAnon ){
+    printf("Cache-control: public\r\n");
+  }else{
+    /* Content intended for logged in users should only be cached in
+    ** the browser, not some shared location.
+    */
+    printf("Cache-control: private\r\n");
+  }
+
 #if CVSTRAC_I18N
-  printf(
-    "Content-Type: %s; charset=%s\r\n"
-    "Content-Length: %d\r\n",
-    zContentType, nl_langinfo(CODESET),
-    nUsedTxt
-  );
+  printf( "Content-Type: %s; charset=%s\r\n", zContentType, nl_langinfo(CODESET));
 #else
-  printf(
-    "Content-Type: %s; charset=ISO-8859-1\r\n"
-    "Content-Length: %d\r\n",
-    zContentType,
-    nUsedTxt
-  );
+  printf( "Content-Type: %s; charset=ISO-8859-1\r\n", zContentType);
 #endif
-  if( fullHttpReply ){
-    printf("Connection: close\r\n");
+
+  if( iReplyStatus != 304 ) {
+    printf( "Content-Length: %d\r\n", nUsedTxt );
   }
   printf("\r\n");
-  if( zTxt ){
+  if( zTxt && iReplyStatus != 304 ){
     fwrite(zTxt, 1, nUsedTxt, stdout);
   }
   if( zLogFile && (log = fopen(zLogFile,"a"))!=0 ){
@@ -240,6 +323,13 @@ void cgi_reply(void){
   }
 }
 
+static int is_same_page(const char *zPage1, const char *zPage2){
+  size_t s1 = strcspn(zPage1,"?#");
+  size_t s2 = strcspn(zPage2,"?#");
+  if( s1 != s2 ) return 0;
+  return !strncmp(zPage1,zPage2,s1);
+}
+
 /*
 ** Do a redirect request to the URL given in the argument.
 **
@@ -252,14 +342,30 @@ void cgi_redirect(const char *zURL){
   if( strncmp(zURL,"http:",5)==0 || strncmp(zURL,"https:",6)==0 || *zURL=='/' ){
     /* An absolute URL.  Do nothing */
   }else{
-    int i, j;
+    int i, j, k=0;
     char *zDest;
     char *zCur = getenv("REQUEST_URI");
     if( zCur==0 ) zCur = "";
     for(i=0; zCur[i] && zCur[i]!='?' && zCur[i]!='#'; i++){}
+    if( g.zExtra ){
+      /* Skip to start of extra stuff, then pass over any /'s that might
+      ** have separated the document root from the extra stuff. This
+      ** ensures that the redirection actually redirects the root, not
+      ** something deep down at the bottom of a URL.
+      */
+      i -= strlen(g.zExtra);
+      while( i>0 && zCur[i-1]=='/' ){ i--; }
+    }
     while( i>0 && zCur[i-1]!='/' ){ i--; }
     zDest = mprintf("%.*s/%s", i, zCur, zURL);
-    for(i=j=0; zDest[i]; i++){
+    /* don't touch the protocol stuff, if it exists */
+    if( !strncmp(zDest,"http://",7) ){
+      k = 7;
+    }else if( !strncmp(zDest,"https://",8) ){
+      k = 8;
+    }
+    /* strip out constructs like .., /./, //, etc */
+    for(i=j=k; zDest[i]; i++){
       if( zDest[i]=='/' ){
         if( zDest[i+1]=='.' ){
           if( zDest[i+2]=='/' ){
@@ -276,12 +382,23 @@ void cgi_redirect(const char *zURL){
             continue;
           }
         }
-        if( zDest[i+1]=='/' ) continue;
+        if( zDest[i+1]=='/' && (i==0 || zDest[i-1]!=':') ) continue;
       }
       zDest[j++] = zDest[i];
     }
     zDest[j] = 0;
     zURL = zDest;
+
+    /* see if we've got a cycle by matching everything up to the ? or #
+    ** in the new and old URLs.
+    */
+    if( is_same_page(zDest,zCur) ){
+      cgi_reset_content();
+      cgi_printf("<html>\n<p>Cyclic redirection in %s</p>\n</html>\n", zURL);
+      cgi_set_status(500, "Internal Server Error");
+      cgi_reply();
+      exit(0);
+    }
   }
 
 /*
@@ -313,7 +430,7 @@ void cgi_redirect(const char *zURL){
   zLocation = mprintf("Location: %s\r\n", zURL);
   cgi_append_header(zLocation);
   cgi_reset_content();
-  cgi_printf("<html>\n<p>Redirect to %s</p>\n</html>\n", zURL);
+  cgi_printf("<html>\n<p>Redirect to %h</p>\n</html>\n", zURL);
   cgi_set_status(302, "Moved Temporarily");
   free(zLocation);
   cgi_reply();
@@ -698,7 +815,7 @@ void cgi_print_all(void){
 /*
 ** Write HTML text for an option menu to standard output.  zParam
 ** is the query parameter that the option menu sets.  zDflt is the
-** initial value of the option menu.  Addition arguments are name/value
+** initial value of the option menu.  Additional arguments are name/value
 ** pairs that define values on the menu.  The list is terminated with
 ** a single NULL argument.
 */
@@ -830,6 +947,372 @@ void cgi_v_optionmenu2(
   cgi_printf("%*s</select>\n", in, "");
 }
 
+/*
+** This routine should never be called directly. Use wrapper functions below.
+** Generates HTML input element to be used in forms.
+** Parameters are explained below inline. If any param is 0 that 
+** attribute/feature will not be used. zValue is required, except for text
+** fields. zName is also required except for submit, reset and button.
+*/
+void cgi_input_elem(
+  int nType,              /* 1:submit, 2:reset, 3:button, 4:file, 5:hidden,
+                          ** 6:checkbox, 7:radio, 8:password, 9:text */
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  int nSize,              /* Used only for text fields */
+  int nMaxLen,            /* Used only for text fields */
+  int nLabelOnLeft,       /* If set, put label text left of element */
+  const char *zValue,     /* Element's value */
+  const char *zDflt,      /* If same as zValue, "select" this element */
+  const char *zLabel      /* Label text. No HTML escaping is done on it */
+){
+  /* Buttons and hidden fields can't have label
+  */
+  int bHasLabel = ( nType>4 && zLabel && zLabel[0] );
+
+  assert( nType > 0 );
+  assert( nType <= 9 );
+  if( zValue==0 ) return;
+  if( nType<1 && nType>3 && (!zName || !zName[0]) ) return;
+  if( bHasLabel ){
+    /* Make sure we have some valid id because <label> won't work in IE w/o it
+    */
+    if( !zId || !zId[0] ) zId = mprintf("%s%h", zName, zValue);
+    @ <label for="%h(zId)">\
+    if( nLabelOnLeft>0 ){
+      @ %s(zLabel) \
+    }
+  }
+  
+  @ <input\
+  switch( nType ){
+    case 1: cgi_printf(" type=\"submit\""); break;
+    case 2: cgi_printf(" type=\"reset\""); break;
+    case 3: cgi_printf(" type=\"button\""); break;
+    case 4: cgi_printf(" type=\"file\""); break;
+    case 5: cgi_printf(" type=\"hidden\""); break;
+    case 6: cgi_printf(" type=\"checkbox\""); break;
+    case 7: cgi_printf(" type=\"radio\""); break;
+    case 8: cgi_printf(" type=\"password\""); break;
+    case 9:
+      @  type="text"\
+      if( nSize>0 ){
+        @  size="%d(nSize)"\
+      }
+      if( nMaxLen>0 ){
+        @  maxlength="%d(nMaxLen)"\
+      }
+      break;
+    default: return;
+  }
+  
+  if( zName && zName[0] ){
+    @  name="%h(zName)"\
+  }
+  if( zId && zId[0] ){
+    @  id="%h(zId)"\
+  }
+  if( zClass && zClass[0] ){
+    @  class="%h(zClass)"\
+  }
+  if( isalnum(nAccessKey) ){
+    @  accesskey="%c(nAccessKey)"\
+  }
+  if( nTabIndex>0 ){
+    @  tabindex="%d(nTabIndex)"\
+  }
+  if( zValue && zValue[0] ){
+    @  value="%h(zValue)"\
+  }
+  if( zDflt && zDflt[0] && strcmp(zDflt, zValue)==0 ){
+    @  checked\
+  }  
+  @ >\
+  
+  if( bHasLabel ){
+    if( nLabelOnLeft<=0 ){
+      @  %s(zLabel)\
+    }
+    @ </label>
+  }else{
+    @ 
+  }
+}
+
+void cgi_submit(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  const char *zValue      /* Element's value */
+){
+  cgi_input_elem(
+    1, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, 0, zValue, 0, 0
+  );
+}
+
+void cgi_reset(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  const char *zValue      /* Element's value */
+){
+  cgi_input_elem(
+    2, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, 0, zValue, 0, 0
+  );
+}
+
+void cgi_button(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  const char *zValue      /* Element's value */
+){
+  cgi_input_elem(
+    3, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, 0, zValue, 0, 0
+  );
+}
+
+void cgi_file(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  const char *zValue      /* Element's value */
+){
+  cgi_input_elem(
+    4, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, 0, zValue, 0, 0
+  );
+}
+
+
+void cgi_hidden(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zValue      /* Element's value */
+){
+  cgi_input_elem(5, zName, zId, 0, 0, 0, 0, 0, 0, zValue, 0, 0);
+}
+
+void cgi_checkbox(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  int nLabelOnLeft,       /* If set, put label text left of element */
+  const char *zValue,     /* Element's value */
+  const char *zDflt,      /* If same as zValue, "select" this element */
+  const char *zLabel      /* Label text. No HTML escaping is done on it */
+){
+  cgi_input_elem(
+    6, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, nLabelOnLeft,
+    zValue, zDflt, zLabel
+  );
+}
+
+void cgi_radio(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  int nLabelOnLeft,       /* If set, put label text left of element */
+  const char *zValue,     /* Element's value */
+  const char *zDflt,      /* If same as zValue, "select" this element */
+  const char *zLabel      /* Label text. No HTML escaping is done on it */
+){
+  cgi_input_elem(
+    7, zName, zId, zClass, nAccessKey, nTabIndex, 0, 0, nLabelOnLeft,
+    zValue, zDflt, zLabel
+  );
+}
+
+void cgi_password(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  int nSize,              /* Field size */
+  int nMaxLen,            /* Maximum number of chars field will accept */
+  int nLabelOnLeft,       /* If set, put label text left of element */
+  const char *zValue,     /* Element's value */
+  const char *zLabel      /* Label text. No HTML escaping is done on it */
+){
+  cgi_input_elem(
+    8, zName, zId, zClass, nAccessKey, nTabIndex, nSize, nMaxLen,
+    nLabelOnLeft, zValue, 0, zLabel
+  );
+}
+
+void cgi_text(
+  const char *zName,      /* CGI param name */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  int nSize,              /* Field size */
+  int nMaxLen,            /* Maximum number of chars field will accept */
+  int nLabelOnLeft,       /* If set, put label text left of element */
+  const char *zValue,     /* Element's value */
+  const char *zLabel      /* Label text. No HTML escaping is done on it */
+){
+  cgi_input_elem(
+    9, zName, zId, zClass, nAccessKey, nTabIndex, nSize, nMaxLen,
+    nLabelOnLeft, zValue, 0, zLabel
+  );
+}
+
+/*
+** Generates radio button elements grouped in <fieldset>.
+** Parameters are explained below inline. If any param is 0 that 
+** attribute/feature will not be used. zName and at least one
+** accesskey/value/label trio are required.
+*/
+void cgi_radio_fieldset(
+  const char *zTitle,     /* Title of <fieldset>. If 0 <fieldset> is not used */
+  const char *zName,      /* CGI param name */
+  const char *zClass,     /* CSS class to apply to fieldset */
+  int *nTabIndex,         /* This will be incremented for each element. 
+                          ** Calling func should be able to continue with it */
+  const char *zDflt,      /* Element with this value will be checked */
+  ...                     /* NULL-terminated list of value/accesskey/label trios */
+){
+  char *zVal, *zLabel;
+  char nAccessKey;
+  va_list ap;
+  if( !zName || !zName[0] ) return;
+  
+  if( zTitle && zTitle[0] ){
+    @ <fieldset\
+    if( zClass && zClass[0] ){
+      @  class="%h(zClass)"\
+    }
+    @ ><legend>%s(zTitle)</legend>
+  }
+
+  /*
+  ** args are: accesskey, value, label
+  */
+  va_start(ap, zDflt);
+  while( (zVal = va_arg(ap, char*))!=0 ){
+    nAccessKey = va_arg(ap, int);
+    zLabel = va_arg(ap, char*);
+    if( zLabel==0 ) break;
+
+    cgi_radio(zName, 0, 0, isprint(nAccessKey)? nAccessKey : 0,
+              nTabIndex==0 ? 0 : ++(*nTabIndex), 0,
+              zVal, zDflt, zLabel);
+  }
+  va_end(ap);
+  if( nTabIndex ) (*nTabIndex)++;
+  if( zTitle && zTitle[0] ){
+    @ </fieldset>
+  }
+}
+
+/*
+** Generates checkbox elements grouped in <fieldset>.
+** Parameters are explained below inline. If any param is 0 that 
+** attribute/feature will not be used. At least one quintet of
+** name/accesskey/value/default value/label is required.
+** If value and default value are equal, element will be "checked".
+*/
+void cgi_checkbox_fieldset(
+  const char *zTitle,     /* Title of <fieldset> */
+  const char *zClass,     /* CSS class to apply to fieldset */
+  int *nTabIndex,         /* This will be incremented for each element. 
+                          ** Calling func should be able to continue with it */
+  ...                     /* NULL-terminated list of
+                          ** name/value/accesskey/label trios */
+){
+  va_list ap;
+  int nAccessKey;
+  char *zName, *zVal, *zDflt, *zLabel;
+  
+  if( zTitle && zTitle[0] ){
+    @ <fieldset\
+    if( zClass && zClass[0] ){
+      @  class="%h(zClass)"\
+    }
+    @ ><legend>%s(zTitle)</legend>
+  }
+  va_start(ap, nTabIndex);
+  while( (zName = va_arg(ap, char*))!=0 ){
+    zVal = va_arg(ap, char*);
+    nAccessKey = va_arg(ap, int);
+    zDflt = va_arg(ap, char*);
+    zLabel = va_arg(ap, char*);
+    if( !zName || !zName[0] || !zVal || !zVal[0] || !zLabel || !zLabel[0] ){
+      break;
+    }
+    /* Label won't work without id in IE so we dummy up id here.
+    ** Hopefully this won't clash with anyones's CSS.
+    */
+    cgi_checkbox(zName, 0, 0, nAccessKey, nTabIndex==0 ? 0 : ++(*nTabIndex), 0,
+              zVal, zDflt, zLabel);
+  }
+  va_end(ap);
+  if( nTabIndex ) (*nTabIndex)++;
+  if( zTitle && zTitle[0] ){
+    @ </fieldset>
+  }
+}
+
+/*
+** Generates HTML links (<a> element).
+** Parameters are explained below inline. If any param is 0 that 
+** attribute/feature will not be used. zText and zHref are required.
+*/
+void cgi_href(
+  const char *zText,      /* Link text */
+  const char *zId,        /* HTML element id */
+  const char *zClass,     /* CSS class to apply */
+  char nAccessKey,        /* Access key to assign */
+  int nTabIndex,          /* Element's tab index */
+  const char *zTitle,     /* Title to apply to <a> */
+  const char *zHref,      /* Link address (vxprintf() format) */
+  ...
+){
+  va_list ap;
+  
+  if( !zText || !zText[0] || !zHref || !zHref[0] ) return;
+  
+  @ <a href="\
+  va_start(ap, zHref);
+  cgi_vprintf(zHref, ap);
+  va_end(ap);
+  @ "\
+  
+  if( zId && zId[0] ){
+    @  id="%h(zId)"\
+  }
+  if( zClass && zClass[0] ){
+    @  class="%h(zClass)"\
+  }
+  if( isalnum(nAccessKey) ){
+    @  accesskey="%c(nAccessKey)"\
+  }
+  if( nTabIndex>0 ){
+    @  tabindex="%d(nTabIndex)"\
+  }
+  if( zTitle && zTitle[0] ){
+    @  title="%h(zTitle)"\
+  }
+  
+  @ >%h(zText)</a>
+}
+
 /*
 ** The "printf" code that follows dates from the 1980's.  It is in
 ** the public domain.  The original comments are included here for
@@ -931,8 +1414,8 @@ static et_info fmtinfo[] = {
   { 's',   0,  0,                  0,    0, etSTRING,     }, 
   { 'z',   0,  0,                  0,    0, etDYNAMIC,    }, 
   { 'h',   0,  0,                  0,    0, etHTMLIZE,    },
-  { 't',   0,  0,                  0,    0, etHTTPIZE,    },
-  { 'T',   0,  0,                  0,    0, etURLIZE,     },
+  { 't',   0,  0,                  0,    0, etHTTPIZE,    }, /* / -> %2F */
+  { 'T',   0,  0,                  0,    0, etURLIZE,     }, /* / -> / */
   { 'c',   0,  0,                  0,    0, etCHARX,      },
   { 'o',   8,  "01234567",         0,  "0", etRADIX,      },
   { 'u',  10,  "0123456789",       0,    0, etRADIX,      },
@@ -1571,6 +2054,76 @@ char *vmprintf(const char *zFormat, va_list ap){
   return sMprintf.zText;
 }
 
+/* 
+** This function implements the callback from vxprintf. 
+**
+** This routine add nNewChar characters of text in zNewText to
+** the sgMprintf structure pointed to by "arg". Unlink mout(), it
+** doesn't grow the buffer, but truncates the output.
+*/
+static void bout(void *arg, char *zNewText, int nNewChar){
+  struct sgMprintf *pM = (struct sgMprintf*)arg;
+  if( pM->nChar + nNewChar + 1 <= pM->nAlloc ){
+    memcpy(&pM->zText[pM->nChar], zNewText, nNewChar);
+    pM->zText[pM->nChar+nNewChar] = 0;
+  }
+  pM->nChar += nNewChar;
+}
+
+/*
+** bprintf() works like snprintf(), but uses the more advanced formatting
+** of vxprintf(). Returns the number of bytes needed to write the full
+** formatted string. Formatted buffer is always NUL terminated.
+*/
+int bprintf(char* zBuf, int nBuflen, const char *zFormat, ...){
+  va_list ap;
+  struct sgMprintf sMprintf;
+
+  if( nBuflen <= 0 ) return -1;
+
+  sMprintf.nChar = 0;
+  sMprintf.nAlloc = nBuflen;
+  sMprintf.zText = zBuf;
+
+  /* bout() NUL terminates... _if_ it's called */
+  sMprintf.zText[0] = 0;
+
+  va_start(ap,zFormat);
+  vxprintf(bout,&sMprintf,zFormat,ap);
+  va_end(ap);
+
+  return sMprintf.nChar;
+}
+
+/*
+** appendf() is basically an accumulating version of bprintf. Each call
+** will append a NUL terminated to the previous output. If nCurlen is
+** NULL, the function will calculate the current length of the buffer itself
+** (but it _does_ assume that it's NUL terminated).
+*/
+int appendf(char* zBuf, int* nCurlen, int nBuflen, const char *zFormat, ...){
+  va_list ap;
+  struct sgMprintf sMprintf;
+
+  if( nBuflen <= 0 ) return -1;
+
+  sMprintf.nChar = nCurlen ? *nCurlen : strlen(zBuf);
+  sMprintf.nAlloc = nBuflen;
+  sMprintf.zText = zBuf;
+
+  if( sMprintf.nChar<nBuflen ){
+    /* bout() NUL terminates... _if_ it's called */
+    sMprintf.zText[sMprintf.nChar] = 0;
+  }
+
+  va_start(ap,zFormat);
+  vxprintf(bout,&sMprintf,zFormat,ap);
+  va_end(ap);
+
+  if( nCurlen ) *nCurlen = sMprintf.nChar;
+  return sMprintf.nChar;
+}
+
 /* 
 ** This function implements the callback from vxprintf. 
 **
@@ -1793,14 +2346,15 @@ static unsigned char zBase[] =
 **
 ** Space to hold the result comes from malloc().
 */
-char *encode64(const char *zData, int nData){
-  char *z64;
+unsigned char *encode64(const unsigned char *zData, int nData){
+  unsigned char *z64;
   int i, n;
 
   if( nData<=0 ){
-    nData = strlen(zData);
+    nData = strlen((const char *)zData);
   }
   z64 = malloc( (nData*4)/3 + 6 );
+  if(z64==0) return 0;
   for(i=n=0; i+2<nData; i+=3){
     z64[n++] = zBase[ (zData[i]>>2) & 0x3f ];
     z64[n++] = zBase[ ((zData[i]<<4) & 0x30) | ((zData[i+1]>>4) & 0x0f) ];
@@ -1826,8 +2380,8 @@ char *encode64(const char *zData, int nData){
 **
 ** Space to hold the decoded string is obtained from malloc().
 */
-char *decode64(const char *z64){
-  char *zData;
+unsigned char *decode64(const unsigned char *z64){
+  unsigned char *zData;
   int n64;
   int i, j;
   int a, b, c, d;
@@ -1839,9 +2393,10 @@ char *decode64(const char *z64){
     for(i=0; zBase[i]; i++){ trans[zBase[i] & 0x7f] = i; }
     isInit = 1;
   }
-  n64 = strlen(z64);
+  n64 = strlen((const char *)z64);
   while( n64>0 && z64[n64-1]=='=' ) n64--;
   zData = malloc( (n64*3)/4 + 4 );
+  if( zData==0 ) return 0;
   for(i=j=0; i+3<n64; i+=4){
     a = trans[z64[i] & 0x7f];
     b = trans[z64[i+1] & 0x7f];
@@ -1909,7 +2464,7 @@ void cgi_handle_http_request(void){
   char *z, *zToken;
   int i;
   struct sockaddr_in remoteName;
-  int size = sizeof(struct sockaddr_in);
+  size_t size = sizeof(struct sockaddr_in);
   char zLine[2000];     /* A single line of input. */
 
   fullHttpReply = 1;
@@ -1964,6 +2519,10 @@ void cgi_handle_http_request(void){
       putenv(mprintf("CONTENT_TYPE=%s", zVal));
     }else if( strcmp(zFieldName,"cookie:")==0 ){
       putenv(mprintf("HTTP_COOKIE=%s", zVal));
+    }else if( strcmp(zFieldName,"if-none-match:")==0 ){
+      putenv(mprintf("HTTP_IF_NONE_MATCH=%s", zVal));
+    }else if( strcmp(zFieldName,"if-modified-since:")==0 ){
+      putenv(mprintf("HTTP_IF_MODIFIED_SINCE=%s", zVal));
     }
   }
 
@@ -1983,11 +2542,12 @@ void cgi_http_server(int iPort){
   int listener;                /* The server socket */
   int connection;              /* A socket for each individual connection */
   fd_set readfds;              /* Set of file descriptors for select() */
-  int lenaddr;                 /* Length of the inaddr structure */
+  size_t lenaddr;              /* Length of the inaddr structure */
   int child;                   /* PID of the child process */
   int nchildren = 0;           /* Number of child processes */
   struct timeval delay;        /* How long to wait inside select() */
   struct sockaddr_in inaddr;   /* The socket address */
+  int opt = 1;                 /* setsockopt flag */
 
   memset(&inaddr, 0, sizeof(inaddr));
   inaddr.sin_family = AF_INET;
@@ -1998,6 +2558,10 @@ void cgi_http_server(int iPort){
     fprintf(stderr,"Can't create a socket\n");
     exit(1);
   }
+
+  /* if we can't terminate nicely, at least allow the socket to be reused */
+  setsockopt(listener,SOL_SOCKET,SO_REUSEADDR,&opt,sizeof(opt));
+
   if( bind(listener, (struct sockaddr*)&inaddr, sizeof(inaddr))<0 ){
     fprintf(stderr,"Can't bind to port %d\n", iPort);
     exit(1);
@@ -2040,3 +2604,68 @@ void cgi_http_server(int iPort){
   /* NOT REACHED */  
   exit(1);
 }
+
+/*
+** Returns an RFC822-formatted time string suitable for HTTP headers, among
+** other things.
+** Returned timezone is always GMT as required by HTTP/1.1 specification.
+**
+** See http://www.faqs.org/rfcs/rfc822.html, section 5
+** and http://www.faqs.org/rfcs/rfc2616.html, section 3.3.
+*/
+char *cgi_rfc822_datestamp(time_t now){
+  static char *azDays[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", 0};
+  static char *azMonths[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
+                             "Aug", "Sep", "Oct", "Nov", "Dec", 0};
+  struct tm *pTm;
+  pTm = gmtime(&now);
+  if( pTm==0 ) return "";
+  return mprintf("%s, %d %s %02d %02d:%02d:%02d GMT",
+                 azDays[pTm->tm_wday], pTm->tm_mday, azMonths[pTm->tm_mon],
+                 pTm->tm_year+1900, pTm->tm_hour, pTm->tm_min, pTm->tm_sec);
+}
+
+/*
+** Parse an RFC822-formatted timestamp as we'd expect from HTTP and return
+** a Unix epoch time. <= zero is returned on failure.
+**
+** Note that this won't handle all the _allowed_ HTTP formats, just the
+** most popular one (the one generated by cgi_rfc822_datestamp(), actually).
+*/
+time_t cgi_rfc822_parsedate(const char *zDate){
+  static char *azMonths[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
+                             "Aug", "Sep", "Oct", "Nov", "Dec", 0};
+  struct tm t;
+  char zIgnore[16];
+  char zMonth[16];
+
+  memset(&t, 0, sizeof(t));
+  if( 7==sscanf(zDate, "%12[A-Za-z,] %d %12[A-Za-z] %d %d:%d:%d", zIgnore,
+                       &t.tm_mday, zMonth, &t.tm_year, &t.tm_hour, &t.tm_min,
+                       &t.tm_sec)){
+
+    if( t.tm_year > 1900 ) t.tm_year -= 1900;
+    for(t.tm_mon=0; azMonths[t.tm_mon]; t.tm_mon++){
+      if( !strncasecmp( azMonths[t.tm_mon], zMonth, 3 )){
+        return mkgmtime(&t);
+      }
+    }
+  }
+
+  return 0;
+}
+
+/*
+** Check the objectTime against the If-Modified-Since request header. If the
+** object time isn't any newer than the header, we immediately send back
+** a 304 reply and exit.
+*/
+void cgi_modified_since(time_t objectTime){
+  const char *zIf = getenv("HTTP_IF_MODIFIED_SINCE");
+  if( zIf==0 ) return;
+  if( objectTime > cgi_rfc822_parsedate(zIf) ) return;
+  cgi_set_status(304,"Not Modified");
+  cgi_reset_content();
+  cgi_reply();
+  exit(0);
+}
