@@ -1,4 +1,4 @@
-static const char ident[] = "@(#) $Header: /cvstrac/cvstrac/makeheaders.c,v 1.3 2003/05/13 01:13:16 drh Exp $";
+static const char ident[] = "@(#) $Header: /cvstrac/cvstrac/makeheaders.c,v 1.4.10.2 2006/12/14 00:35:46 cpb Exp $";
 /*
 ** This program scans C and C++ source files and automatically generates
 ** appropriate header files.
@@ -9,10 +9,9 @@ static const char ident[] = "@(#) $Header: /cvstrac/cvstrac/makeheaders.c,v 1.3
 #include <memory.h>
 #include <sys/stat.h>
 #include <assert.h>
+#include <string.h>
 #ifndef WIN32
 # include <unistd.h>
-#else
-# include <string.h>
 #endif
 
 /*
@@ -2535,7 +2534,7 @@ static void DeclareObject(
   int isCpp;             /* True if generating C++ */
   int doneTypedef = 0;   /* True if a typedef has been done for this object */
 
-  // printf("BEGIN %s of %s\n",needFullDecl?"FULL":"PROTOTYPE",pDecl->zName);
+  /* printf("BEGIN %s of %s\n",needFullDecl?"FULL":"PROTOTYPE",pDecl->zName);*/
   /* 
   ** For any object that has a forward declaration, go ahead and do the
   ** forward declaration first.
@@ -2979,7 +2978,7 @@ static void DocumentationDump(void){
 ** documentation record for the header comment at the beginning of the
 ** file (if the file has a header comment.)
 */
-void PrintModuleRecord(const char *zFile, const char *zFilename){
+static void PrintModuleRecord(const char *zFile, const char *zFilename){
   int i;
   static int addr = 5;
   while( isspace(*zFile) ){ zFile++; }
