@@ -12,7 +12,7 @@ gzip -f cvstrac-$VERS.bin
 DIR=`pwd`
 cd $SRCDIR/..
 mv cvstrac cvstrac-$VERS
-tar czf $DIR/cvstrac-$VERS.tar.gz `find cvstrac-$VERS | grep -v CVS`
+tar czf $DIR/cvstrac-$VERS.tar.gz `find cvstrac-$VERS -type f | grep -v CVS`
 mv cvstrac-$VERS cvstrac
 cd $DIR
 echo 'rsync to www.hwaci.com'
